# -----------------------------------------------------------------------
# Copyright (c) 2021    Cornell Database Group
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
# IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
# -----------------------------------------------------------------------

import logging
import time

import MySQLdb

from .abstractdriver import *

class MysqlDriver(AbstractDriver):
    """the DBMS driver for MySQL"""

    def __init__(self, conf, sys_params):
        super(MysqlDriver, self).__init__("mysql", conf, sys_params)

    def connect(self):
        """connect to a database"""
        self.conn = MySQLdb.connect(self.config['host'], self.config['user'], self.config['passwd'], self.config['db'])
        self.cursor = self.conn.cursor()
        self.index_creation_format = "CREATE INDEX %s ON %s (%s) USING BTREE;"
        self.index_drop_format = "ALTER TABLE %s drop index %s;"
        # self.is_cluster = True
        self.sys_params_type = len(self.sys_params)
        self.sys_params_space = [len(specific_parameter) for specific_parameter in self.sys_params]
        self.retrieve_table_name_sql = "show tables;"
        self.cardinality_format = "select count(*) from %s;"

    def cardinalities(self):
        """get cardinality of the connected database"""
        self.cursor.execute(self.retrieve_table_name_sql)
        dbms_tables = []
        cardinality_info = {}
        for (table_name,) in self.cursor:
            dbms_tables.append(table_name)
        for table in dbms_tables:
            self.cursor.execute(self.cardinality_format % table)
            result = self.cursor.fetchone()
            cardinality = result[0]
            cardinality_info[table] = cardinality
        return cardinality_info

    def run_queries_with_timeout(self, query_list, timeout):
        """run queries with a timeout"""
        run_time = []
        for query_sql, current_timeout in zip(query_list, timeout):
            try:
                current_timeout = current_timeout
                self.cursor.execute("SET SESSION MAX_EXECUTION_TIME=%d" % (current_timeout * 1000))
                start_time = time.time()
                self.cursor.execute(query_sql)
                finish_time = time.time()
                duration = finish_time - start_time
                print("query run:%s" % duration)
            except MySQLdb.OperationalError as oe:
                # print(oe)
                # print("timeout")
                duration = current_timeout
            run_time.append(duration)
        print(run_time)
        # reset back to default configuraiton
        try:
            self.cursor.execute("SET SESSION MAX_EXECUTION_TIME=0")
            self.cursor.execute("drop view if exists REVENUE0;")
        except MySQLdb.OperationalError as oe:
            print(oe)
        return run_time

    def run_queries_without_timeout(self, query_list):
        """run queries without timeout"""
        return self.run_queries_with_timeout(query_list, [0 for query in query_list])

    def build_index(self, index_to_create):
        """build index"""
        index_sql = self.index_creation_format % (index_to_create[0], index_to_create[1], index_to_create[2])
        logging.debug(f"create index {index_sql}")
        self.cursor.execute(index_sql)
        self.conn.commit()

    def build_index_command(self, index_to_create):
        """build index command"""
        index_sql = self.index_creation_format % (index_to_create[0], index_to_create[1], index_to_create[2])
        return index_sql

    def drop_index(self, index_to_drop):
        """drop index"""
        index_sql = self.index_drop_format % (index_to_drop[1], index_to_drop[0])
        logging.debug(f"drop index {index_sql}")
        self.cursor.execute(index_sql)
        self.conn.commit()

    def set_system_parameter(self, parameter_sql):
        """switch system parameters"""
        self.cursor.execute(parameter_sql)
        self.conn.commit()

## CLASS



# -----------------------------------------------------------------------
# Copyright (c) 2021    Cornell Database Group
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
# IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
# -----------------------------------------------------------------------

import logging
import time

import psycopg2
from psycopg2._psycopg import InternalError
from psycopg2._psycopg import QueryCanceledError

from .abstractdriver import *

class PostgresDriver(AbstractDriver):
    """the DBMS driver for Postgres"""

    def __init__(self, conf, sys_params):
        super(PostgresDriver, self).__init__("postgres", conf, sys_params)

    def connect(self):
        """connect to a database"""
        self.conn = psycopg2.connect("dbname='%s' user='%s'" % (self.config["db"], self.config["user"]))
        self.conn.autocommit = True
        self.cursor = self.conn.cursor()
        self.index_creation_format = "CREATE INDEX %s ON %s (%s);"
        self.index_drop_format = "drop index %s;"
        self.is_cluster = False
        self.retrieve_table_name_sql = "SELECT table_name FROM information_schema.tables WHERE table_schema = 'public' ORDER BY table_name;"
        self.cardinality_format = "select count(*) from %s;"
        self.cluster_indices_format = "CLUSTER %s ON %s;"

    def cardinalities(self):
        """get cardinality of the connected database"""
        self.cursor.execute(self.retrieve_table_name_sql)
        dbms_tables = []
        cardinality_info = {}
        for table in self.cursor.fetchall():
            dbms_tables.append(table)
        for table in dbms_tables:
            self.cursor.execute(self.cardinality_format % table)
            result = self.cursor.fetchone()
            cardinality = result[0]
            cardinality_info[table[0].lower()] = cardinality
        return cardinality_info

    def run_queries_with_timeout(self, query_list, timeout):
        """run queries with a timeout"""
        run_time = []
        for query_sql, current_timeout in zip(query_list, timeout):
            try:
                logging.debug(f"query sql: {query_sql}")
                logging.debug(f"current timeout: {current_timeout}")
                current_timeout = current_timeout
                self.cursor.execute("set statement_timeout = %d" % (current_timeout * 1000))
                start_time = time.time()
                self.cursor.execute(query_sql)
                finish_time = time.time()
                duration = finish_time - start_time
            except QueryCanceledError:
                duration = current_timeout
            except InternalError:
                # error to run the query, set duration to a large number
                logging.debug(f"Internal Error for query {query_sql}")
                duration = current_timeout * 1000
            logging.debug(f"duration: {duration}")
            run_time.append(duration)
        # reset the timeout to the default configuration
        self.cursor.execute("set statement_timeout=0;")
        self.cursor.execute("drop view if exists REVENUE0;")
        return run_time

    def run_queries_without_timeout(self, query_list):
        """run queries without timeout"""
        return self.run_queries_with_timeout(query_list, [0] * len(query_list))

    def build_index(self, index_to_create):
        """build index"""
        index_sql = self.index_creation_format % (index_to_create[0], index_to_create[1], index_to_create[2])
        logging.debug(f"create index {index_sql}")
        self.cursor.execute(index_sql)
        # if we consider the cluster indices
        if self.is_cluster:
            self.cursor.execute(self.cluster_indices_format % (index_to_create[0], index_to_create[1]))
        # self.conn.commit()

    def build_index_command(self, index_to_create):
        """build index command"""
        index_sql = self.index_creation_format % (index_to_create[0], index_to_create[1], index_to_create[2])
        if self.is_cluster:
            cluster_sql = self.cluster_indices_format % (index_to_create[0], index_to_create[1])
            return f"{index_sql} \n {cluster_sql}"
        else:
            return index_sql

    def drop_index(self, index_to_drop):
        """drop index"""
        index_sql = self.index_drop_format % (index_to_drop[0])
        logging.debug(f"drop index {index_sql}")
        self.cursor.execute(index_sql)
        # self.conn.commit()

    def set_system_parameter(self, parameter_sql):
        """switch system parameters"""
        self.cursor.execute(parameter_sql)
        # self.conn.commit()

## CLASS



#!/usr/bin/env python3
# -----------------------------------------------------------------------
# Copyright (c) 2021    Cornell Database Group
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
# IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
# -----------------------------------------------------------------------

import argparse
import json
import logging
import os
import random
import sys

import numpy as np

from udo.agent.ddpg_agent import run_ddpg_agent
from udo.agent.sarsa_agent import run_sarsa_agent
from udo.agent.udo_agent import run_udo_agent
from udo.agent.udo_simplifed_agent import run_simplifed_udo_agent
from udo.drivers.mysqldriver import MysqlDriver
from udo.drivers.postgresdriver import PostgresDriver

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO,
                        format="%(asctime)s [%(funcName)s:%(lineno)03d] %(levelname)-5s: %(message)s",
                        datefmt="%m-%d-%Y %H:%M:%S",
                        stream=sys.stdout)

    udo_parser = argparse.ArgumentParser(description='UDO optimizer.')
    # database setting
    udo_parser.add_argument('-system', choices=('mysql', 'postgres'),
                            help='Target system driver')
    udo_parser.add_argument('-db', default="tpcc",
                            help='the database to optimizes')
    udo_parser.add_argument('-username', default="udo",
                            help='username')
    udo_parser.add_argument('-password', default="udo",
                            help='password')
    udo_parser.add_argument('-queries', default="tpch_query.sql",
                            help='the input query file')
    udo_parser.add_argument('-indices',
                            help='the input query file')
    udo_parser.add_argument('-sys_params',
                            help='the input system params json file')
    # tuning time
    udo_parser.add_argument('-duration', default=5, type=float,
                            help='time for tuning in hours')

    # rl algorithm
    udo_parser.add_argument('-agent', default='udo', choices=('udo', 'udo-s', 'ddpg', 'sarsa'),
                            help='reinforcement learning agent')
    udo_parser.add_argument('-horizon', default=5, type=int,
                            help='the number horizon for reinforcement agent')
    udo_parser.add_argument('-heavy_horizon', default=3, type=int,
                            help='the number horizon for heavy parameters in UDO')
    # mcts algorithm
    udo_parser.add_argument('-rl_update', choices=('RAVE', 'MCTS'), default='RAVE',
                            help='the update policy of UDO tree search')
    udo_parser.add_argument('-rl_select', choices=('UCB1', 'UCBV'), default='UCBV',
                            help='the selection policy of UDO tree search')
    udo_parser.add_argument('-rl_reward', choices=('delta', 'accumulate'),
                            help='the reward of reinforcement learning agent')
    udo_parser.add_argument('-rl_delay', choices=('UCB', 'Exp3'),
                            help='the delay selection policy')
    udo_parser.add_argument('-rl_max_delay_time', type=int, default=5,
                            help='the delay selection policy')

    # tuning configuration
    udo_parser.add_argument('-sample_rate', type=int, default="1",
                            help='sampled rate from workload')
    udo_parser.add_argument('-default_query_time_out', type=int, default="5",
                            help='default timeout in seconds for each query')
    udo_parser.add_argument('-time_out_ratio', type=float, default="1.1",
                            help='timeout ratio respect to default time')

    # load json file
    udo_parser.add_argument('--load_json',
                            help='Load settings from file in json format. Command line options override values in file.')

    args = udo_parser.parse_args()

    # json file the higher priority
    if args.load_json:
        with open(args.load_json, 'rt') as f:
            t_args = argparse.Namespace()
            t_args.__dict__.update(json.load(f))
            args = udo_parser.parse_args(namespace=t_args)

    # change
    args = vars(args)

    # init queries
    if args['queries']:
        queries = dict()
        for file_name in sorted(os.listdir(args['queries'])):
            if file_name.endswith(".sql"):
                with open(os.path.join(args['queries'], file_name)) as f:
                    content = f.read()
                    queries[file_name] = content
        # print(constants.QUERIES)

    # init indices with external files
    if args['indices']:
        # analyze queries to extract indexes
        with open(args['indices']) as f:
            content = f.readlines()
            indices = [x.strip() for x in content]
            candidate_indices = []
            for index_str in indices:
                index_information = index_str.split(";")
                candidate_indices.append((index_information[0], index_information[1], index_information[2]))

    dbms_conf = {
        "host": "127.0.0.1",
        "db": args['db'],
        "user": args['username'],
        "passwd": args['password'],
    }

    # check the validate of configurations file
    if not args['system'] or not args['db']:
        print("Please specific a database")
        exit()

    if not args['duration'] or not args['horizon']:
        print("Wrong parameters. Please check the input parameters")
        exit()

    if not args['sys_params']:
        print("Please specify the system parameters")
        exit()

    with open(args['sys_params'], 'rt') as f:
        sys_params = json.load(f)

    # create a dbms driver
    driver = None
    if args['system'] == "mysql":
        driver = MysqlDriver(dbms_conf, sys_params)
    elif args['system'] == "postgres":
        driver = PostgresDriver(dbms_conf, sys_params)

    # obtain index cardinality information
    driver.connect()
    cardinality_info = driver.cardinalities()

    # obtain index applicable queries
    for i in range(len(candidate_indices)):
        contain_query = []
        for query_id, query_str in queries.items():
            # print(candidate_indices[i][2])
            if "where" in query_str:
                where_clause = query_str[query_str.index("where"):].lower()
            else:
                where_clause = query_str.lower()
            contain_columns = candidate_indices[i][2].lower().split(",")
            if all(contain_column in where_clause for contain_column in contain_columns):
                # if any(contain_column in where_clause for contain_column in contain_columns):
                contain_query.append(query_id)
        index_cardinality = cardinality_info[candidate_indices[i][1].lower()]
        candidate_indices[i] += (contain_query, index_cardinality,)

    # filter indices which contains at least has one appliable query
    candidate_indices = [candidate_index for candidate_index in candidate_indices if len(candidate_index[3]) > 0]

    # print(len(index.candidate_indices))

    # fix the random seed
    random.seed(123)
    np.random.seed(123)

    tuning_config = {
        "duration": args['duration'],
        "horizon": args['horizon'],
        "sample_rate": args['sample_rate'],
        "default_query_time_out": args['default_query_time_out'],
        "time_out_ratio": args['time_out_ratio']
    }

    # if the agent is udo
    if args['agent'] == 'udo':
        if not args['heavy_horizon']:
            print("Please specific the step of heavy configurations")
            exit()
        # run udo
        print(tuning_config)
        horizon = args['horizon']
        tuning_config['heavy_horizon'] = args['heavy_horizon']
        tuning_config['light_horizon'] = horizon - int(args['heavy_horizon'])
        tuning_config['rl_max_delay_time'] = args['rl_max_delay_time']
        tuning_config['rl_update'] = args['rl_update']
        tuning_config['rl_select'] = args['rl_select']
        run_udo_agent(driver=driver, queries=queries, candidate_indices=candidate_indices, tuning_config=tuning_config)
    elif args['agent'] == 'udo-s':
        # run simplified udo
        run_simplifed_udo_agent(driver=driver, queries=queries, candidate_indices=candidate_indices,
                                tuning_config=tuning_config)
    elif args['agent'] == 'ddpg':
        # run ddpg deep rl
        run_ddpg_agent(driver=driver, queries=queries, candidate_indices=candidate_indices, tuning_config=tuning_config)
    elif args['agent'] == 'sarsa':
        # run sarsa deep rl
        run_sarsa_agent(driver=driver, queries=queries, candidate_indices=candidate_indices,
                        tuning_config=tuning_config)

    # print(args.accumulate(args.integers))



from setuptools import setup, find_packages

setup(
    name='udo_optimization',
    version='0.0.2',
    description='gym environment for UDO',
    author='Junxiong Wang',
    author_email='chuangzhetianxia@gmail.coma',
    url='https://ovss.github.io/udo_optimization/',
    install_requires=['gym>=0.17.3'],
    packages=find_packages()
)



import logging
import random
import time

import gym
import udo_optimization

import numpy as np
import tensorflow as tf
from tensorflow.keras import layers


def run_ddpg_agent(driver, queries, candidate_indices, tuning_config):
    """Run DDPG agent for universal optimization"""
    env = gym.make("udo_optimization-v0", driver=driver, queries=queries, candidate_indices=candidate_indices,
                   config=tuning_config)

    num_states = env.observation_space.shape[0]
    num_actions = env.nA_index
    logging.debug(f"Size of State Space ->  {num_states}")
    logging.debug(f"Size of Action Space ->  {num_actions}")

    upper_bound = 1
    lower_bound = -1

    logging.debug(f"Max Value of Action ->  {upper_bound}")
    logging.debug(f"Min Value of Action ->  {lower_bound}")

    # the following code is from Keras RL example, https://github.com/keras-team/keras-io/blob/master/examples/rl/ddpg_pendulum.py
    class OUActionNoise:
        def __init__(self, mean, std_deviation, theta=0.15, dt=1e-2, x_initial=None):
            self.theta = theta
            self.mean = mean
            self.std_dev = std_deviation
            self.dt = dt
            self.x_initial = x_initial
            self.reset()

        def __call__(self):
            # Formula taken from https://www.wikipedia.org/wiki/Ornstein-Uhlenbeck_process.
            x = (
                    self.x_prev
                    + self.theta * (self.mean - self.x_prev) * self.dt
                    + self.std_dev * np.sqrt(self.dt) * np.random.normal(size=self.mean.shape)
            )
            # Store x into x_prev
            # Makes next noise dependent on current one
            self.x_prev = x
            return x

        def reset(self):
            if self.x_initial is not None:
                self.x_prev = self.x_initial
            else:
                self.x_prev = np.zeros_like(self.mean)

    class Buffer:
        def __init__(self, buffer_capacity=100000, batch_size=64):
            # Number of "experiences" to store at max
            self.buffer_capacity = buffer_capacity
            # Num of tuples to train on.
            self.batch_size = batch_size

            # Its tells us num of times record() was called.
            self.buffer_counter = 0

            # Instead of list of tuples as the exp.replay concept go
            # We use different np.arrays for each tuple element
            self.state_buffer = np.zeros((self.buffer_capacity, num_states))
            self.action_buffer = np.zeros((self.buffer_capacity, num_actions))
            self.reward_buffer = np.zeros((self.buffer_capacity, 1))
            self.next_state_buffer = np.zeros((self.buffer_capacity, num_states))

        # Takes (s,a,r,s') obervation tuple as input
        def record(self, obs_tuple):
            # Set index to zero if buffer_capacity is exceeded,
            # replacing old records
            index = self.buffer_counter % self.buffer_capacity

            self.state_buffer[index] = obs_tuple[0]
            self.action_buffer[index] = obs_tuple[1]
            self.reward_buffer[index] = obs_tuple[2]
            self.next_state_buffer[index] = obs_tuple[3]

            self.buffer_counter += 1

        # Eager execution is turned on by default in TensorFlow 2. Decorating with tf.function allows
        # TensorFlow to build a static graph out of the logic and computations in our function.
        # This provides a large speed up for blocks of code that contain many small TensorFlow operations such as this one.
        @tf.function
        def update(
                self, state_batch, action_batch, reward_batch, next_state_batch,
        ):
            # Training and updating Actor & Critic networks.
            # See Pseudo Code.
            with tf.GradientTape() as tape:
                target_actions = target_actor(next_state_batch, training=True)
                y = reward_batch + gamma * target_critic(
                    [next_state_batch, target_actions], training=True
                )
                critic_value = critic_model([state_batch, action_batch], training=True)
                critic_loss = tf.math.reduce_mean(tf.math.square(y - critic_value))

            critic_grad = tape.gradient(critic_loss, critic_model.trainable_variables)
            critic_optimizer.apply_gradients(
                zip(critic_grad, critic_model.trainable_variables)
            )

            with tf.GradientTape() as tape:
                actions = actor_model(state_batch, training=True)
                critic_value = critic_model([state_batch, actions], training=True)
                # Used `-value` as we want to maximize the value given
                # by the critic for our actions
                actor_loss = -tf.math.reduce_mean(critic_value)

            actor_grad = tape.gradient(actor_loss, actor_model.trainable_variables)
            actor_optimizer.apply_gradients(
                zip(actor_grad, actor_model.trainable_variables)
            )

        # We compute the loss and update parameters
        def learn(self):
            # Get sampling range
            record_range = min(self.buffer_counter, self.buffer_capacity)
            # Randomly sample indices
            batch_indices = np.random.choice(record_range, self.batch_size)

            # Convert to tensors
            state_batch = tf.convert_to_tensor(self.state_buffer[batch_indices])
            action_batch = tf.convert_to_tensor(self.action_buffer[batch_indices])
            reward_batch = tf.convert_to_tensor(self.reward_buffer[batch_indices])
            reward_batch = tf.cast(reward_batch, dtype=tf.float32)
            next_state_batch = tf.convert_to_tensor(self.next_state_buffer[batch_indices])

            self.update(state_batch, action_batch, reward_batch, next_state_batch)

    # This update target parameters slowly
    # Based on rate `tau`, which is much less than one.
    @tf.function
    def update_target(target_weights, weights, tau):
        for (a, b) in zip(target_weights, weights):
            a.assign(b * tau + a * (1 - tau))

    def get_actor():
        # Initialize weights between -3e-3 and 3-e3
        last_init = tf.random_uniform_initializer(minval=-0.003, maxval=0.003)

        inputs = layers.Input(shape=(num_states,))
        out = layers.Dense(256, activation="relu")(inputs)
        out = layers.Dense(256, activation="relu")(out)
        outputs = layers.Dense(num_actions, activation="tanh", kernel_initializer=last_init)(out)

        # Our upper bound is 2.0 for Pendulum.
        outputs = outputs * upper_bound
        model = tf.keras.Model(inputs, outputs)
        return model

    def get_critic():
        # State as input
        state_input = layers.Input(shape=(num_states))
        state_out = layers.Dense(16, activation="relu")(state_input)
        state_out = layers.Dense(32, activation="relu")(state_out)

        # Action as input
        action_input = layers.Input(shape=(num_actions))
        action_out = layers.Dense(32, activation="relu")(action_input)

        # Both are passed through seperate layer before concatenating
        concat = layers.Concatenate()([state_out, action_out])

        out = layers.Dense(256, activation="relu")(concat)
        out = layers.Dense(256, activation="relu")(out)
        outputs = layers.Dense(1)(out)

        # Outputs single value for give state-action
        model = tf.keras.Model([state_input, action_input], outputs)

        return model

    def policy(state, noise_object, previous_actions):
        sampled_actions = tf.squeeze(actor_model(state))
        noise = noise_object()
        # Adding noise to action
        sampled_actions = sampled_actions.numpy() + noise
        logging.debug(f"sampled_actions: {sampled_actions}")

        # We make sure action is within bounds
        legal_action = np.clip(sampled_actions, lower_bound, upper_bound)

        action_weights = [np.squeeze(legal_action)]
        print(action_weights)
        max_action_weight = -1000
        nr_action = len(action_weights[0])
        start_action = random.randrange(0, nr_action)
        action = 0
        logging.debug(f"start_action {start_action}")
        # pick up the action which has the highest weight
        for action_idx in range(0, nr_action):
            current_action = (start_action + action_idx) % nr_action
            if (current_action not in previous_actions) and (action_weights[0][current_action] > max_action_weight):
                max_action_weight = action_weights[0][current_action]
                action = current_action

        # action = np.random.choice(np.flatnonzero(action_weights == max(action_weights)))
        # action = np.argmax(action_weights)
        return action

    std_dev = 0.2
    ou_noise = OUActionNoise(mean=np.zeros(1), std_deviation=float(std_dev) * np.ones(1))

    actor_model = get_actor()
    critic_model = get_critic()

    target_actor = get_actor()
    target_critic = get_critic()

    # Making the weights equal initially
    target_actor.set_weights(actor_model.get_weights())
    target_critic.set_weights(critic_model.get_weights())

    # Learning rate for actor-critic models
    critic_lr = 0.002
    actor_lr = 0.001

    critic_optimizer = tf.keras.optimizers.Adam(critic_lr)
    actor_optimizer = tf.keras.optimizers.Adam(actor_lr)

    # Discount factor for future rewards
    gamma = 0.99
    # Used to update target networks
    tau = 0.005

    buffer = Buffer(50000, 64)

    # set the horizon
    env.horizon = tuning_config['horizon']
    duration_in_seconds = tuning_config['duration'] * 3600

    # To store reward history of each episode
    ep_reward_list = []
    # To store average reward history of last few episodes
    avg_reward_list = []

    start_time = time.time()
    current_time = time.time()

    ep = 0
    while (current_time - start_time) < duration_in_seconds:
        # start a new episode
        prev_state = env.reset()
        episodic_reward = 0
        previous_actions = []
        while True:
            # Uncomment this to see the Actor in action
            tf_prev_state = tf.expand_dims(tf.convert_to_tensor(prev_state), 0)
            # select action according to policy
            action = policy(tf_prev_state, ou_noise, previous_actions)
            previous_actions.append(action)
            # receive state and reward from environment.
            state, reward, done, info = env.step(action)

            buffer.record((prev_state, action, reward, state))
            episodic_reward += reward

            current_time = time.time()
            logging.info(f"episode: {ep}")
            logging.info(f"evaluate duration: {(current_time - start_time)}")
            ep += 1

            buffer.learn()
            update_target(target_actor.variables, actor_model.variables, tau)
            update_target(target_critic.variables, critic_model.variables, tau)

            # End this episode when `done` is True
            if done:
                break

            prev_state = state

        ep_reward_list.append(episodic_reward)

        # Mean of last 40 episodes
        avg_reward = np.mean(ep_reward_list[-40:])
        logging.info(f"Episode * {ep} * Avg Reward is ==> {avg_reward}")
        avg_reward_list.append(avg_reward)
        env.print_state_summary(env.best_state)



#!/usr/bin/env python3
# -----------------------------------------------------------------------
# Copyright (c) 2021    Cornell Database Group
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
# IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
# -----------------------------------------------------------------------

import argparse
import json
import os
import re
import itertools

udo_parser = argparse.ArgumentParser(description='UDO index candidate generator.')

udo_parser.add_argument('-db_schema', help='the database schmea to optimizes')
udo_parser.add_argument('-queries', help='queries')

args = udo_parser.parse_args()
# change
args = vars(args)

with open(args['db_schema']) as f:
    db_schema = json.load(f)

if args['queries']:
    queries = dict()
    for file_name in os.listdir(args['queries']):
        if file_name.endswith(".sql"):
            with open(os.path.join(args['queries'], file_name)) as f:
                content = f.read()
                queries[file_name] = content

table_pattern = re.compile("FROM(.*)WHERE")
where_pattern = re.compile("WHERE(.*)")
collect_indices = dict()
for query_id, query in queries.items():
    matches = table_pattern.findall(query)
    join_elements = matches[0].split(',')
    rename_table_dict = {}
    for join_element in join_elements:
        rename_pos = join_element.index("AS")
        if join_element[:rename_pos].strip() not in rename_table_dict:
            rename_table_dict[join_element[:rename_pos].strip()] = [join_element[rename_pos + 2:].strip()]
        else:
            rename_table_dict[join_element[:rename_pos].strip()].append(join_element[rename_pos + 2:].strip())
    print(rename_table_dict)
    where_clause = where_pattern.findall(query)[0]
    for table, columns in db_schema.items():
        if table in rename_table_dict:
            for rename_table in rename_table_dict[table]:
                for column in columns:
                    if rename_table + '.' + column in where_clause:
                        if table not in collect_indices:
                            collect_indices[table] = [column]
                        elif column not in collect_indices[table]:
                            collect_indices[table].append(column)
print(collect_indices)

for k in collect_indices.keys():
    all_candidate = collect_indices[k]
    v = []
    for i in range(1, len(all_candidate) + 1):
        v += [','.join(comb) for comb in (itertools.combinations(all_candidate, i))]
    for i in range(len(v)):
        print("('IDX_%s_%d','%s','%s')," % (k, i, k, v[i]))



# -----------------------------------------------------------------------
# Copyright (c) 2021    Cornell Database Group
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
# IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
# -----------------------------------------------------------------------

import logging
import math
import random
import time

import gym

from udo.mcts.mcts_node import SpaceType
from udo.mcts.uct_node import uct_node
from udo.optimizer import order_optimizer


def run_udo_agent(driver, queries, candidate_indices, tuning_config):
    duration_in_seconds = tuning_config['duration'] * 3600

    env = gym.make('udo_optimization-v0', driver=driver, queries=queries, candidate_indices=candidate_indices,
                   config=tuning_config)

    nr_query = len(queries)  # number of queries
    index_card_info = list(map(lambda x: x[4], candidate_indices))
    index_to_applicable_queries = env.index_to_applicable_queries

    start_tune_time = time.time()
    logging.debug(f"start time: {start_tune_time}")

    optimizer = order_optimizer.OrderOptimizer(index_card_info)

    # number of indices equal heavy_tree_height + 1
    heavy_tree_height = tuning_config['heavy_horizon']
    light_tree_height = tuning_config['light_horizon']
    max_delay_time = tuning_config['rl_max_delay_time']

    init_state = env.state_decoder(0)
    micro_episode = 5

    terminate_action = env.index_candidate_num
    light_tree_cache = dict()
    query_evaluate_sample_rate = 1
    # reset the environment
    env.reset()
    default_runtime = env.default_runtime

    heavy_root = uct_node(round=0, tree_level=0, tree_height=heavy_tree_height, state=init_state, env=env,
                          space_type=SpaceType.Heavy)
    idx_build_time = 0
    best_simulation_time = sum(default_runtime)
    best_configs = {}
    t1 = 1
    start_time = time.time()
    end_episode_time = time.time()
    while (end_episode_time - start_time) < duration_in_seconds:
        selected_heavy_action_batch = []
        configuration_to_evaluate = []
        # remove_terminate_action_batch = []
        logging.debug(f"delay_time: {max_delay_time}")
        for d in range(max_delay_time):
            selected_heavy_actions = heavy_root.sample(t1 + d)
            selected_heavy_action_batch.append(selected_heavy_actions)
            # add all intermediate steps
            remove_terminate_heavy_actions = [heavy_action for heavy_action in selected_heavy_actions if
                                              heavy_action != terminate_action]
            for i in range(len(remove_terminate_heavy_actions)):
                current_actions = frozenset(remove_terminate_heavy_actions[:i + 1])
                if current_actions not in configuration_to_evaluate:
                    configuration_to_evaluate.append(current_actions)
        # show those actions
        logging.debug(f"selected heavy actions: {selected_heavy_action_batch}")
        logging.debug(f"evaluate configurations: {configuration_to_evaluate}")
        # evaluated_order = range(0, len(remove_terminate_action_batch))
        evaluated_order = optimizer.greedy_min_cost_order(configuration_to_evaluate)
        logging.debug(f"evaluated order: {evaluated_order}")
        macro_performance_info = dict()
        for current_order_idx in range(len(evaluated_order)):
            current_order = evaluated_order[current_order_idx]
            selected_heavy_action_frozen = frozenset(configuration_to_evaluate[current_order])
            # generate add action
            add_action = set(selected_heavy_action_frozen)
            drop_action = set()
            if current_order_idx > 0:
                # take the previous created indices
                previous_set = set(configuration_to_evaluate[evaluated_order[current_order_idx - 1]])
            if t1 > 1 or current_order_idx > 0:
                add_action = add_action - previous_set
                drop_action = previous_set - set(selected_heavy_action_frozen)
                logging.debug(f"invoke action: {add_action}")
                logging.debug(f"drop action: {drop_action}")
            # build the indices
            time_start = time.time()
            env.index_step(add_action, drop_action)
            time_end = time.time()
            idx_build_time += (time_end - time_start)

            logging.debug(f"selected_heavy_action_frozen: {selected_heavy_action_frozen}")
            if selected_heavy_action_frozen in light_tree_cache:
                light_root = light_tree_cache[selected_heavy_action_frozen]
            else:
                light_root = uct_node(round=0, tree_level=0, tree_height=light_tree_height, state=init_state, env=env,
                                      space_type=SpaceType.Light)
                light_tree_cache[selected_heavy_action_frozen] = light_root
            # for the light tree
            best_reward = 0
            # query to consider for the current select heavy configuration
            query_to_consider = set(
                [applicable_query for applicable_queries in
                 list(map(lambda x: index_to_applicable_queries[x], selected_heavy_action_frozen)) for applicable_query
                 in
                 applicable_queries])
            logging.debug(f"query to consider: {query_to_consider}")
            # best performance of the sampled each query
            best_micro_performance = dict()
            for t2 in range(1, micro_episode + 1):
                # for the micro episode
                env.reset()
                selected_light_actions = light_root.sample(t1 * micro_episode + t2)
                # evaluate the light actions
                for selected_light_action in selected_light_actions:
                    # move to next state
                    state = env.step_without_evaluation(selected_light_action)
                # obtain sample number
                sample_num = math.ceil(query_evaluate_sample_rate * len(query_to_consider))
                # generate sample queries
                sampled_query_list = random.sample(list(query_to_consider), k=sample_num)
                logging.debug(f"sampled_query_list: {sampled_query_list}")
                # obtain run time info by running queries within timeout
                run_time = env.evaluate(sampled_query_list)
                # the total time of sampled queries
                total_run_time = sum(run_time)
                # the default time of sampled queries
                default_runtime_of_sampled_queries = [default_runtime[select_query] for select_query in
                                                      sampled_query_list]
                sum_default_time_of_sampled_queries = sum(default_runtime_of_sampled_queries)
                # the relative ration of the improvement, the less of total_run_time, the better
                light_reward = sum_default_time_of_sampled_queries / total_run_time
                logging.debug(f"default_runtime_of_sampled_queries {default_runtime_of_sampled_queries}")
                logging.debug(f"light_action: {selected_light_actions}")
                logging.debug(f"light_reward: {light_reward}")

                other_default_time = sum(default_runtime[select_query] for select_query in range(nr_query) if
                                         select_query not in sampled_query_list)
                estimate_workload_time = (other_default_time + total_run_time)
                logging.debug(f"estimate whole workload time: {estimate_workload_time}")
                if estimate_workload_time < best_simulation_time:
                    best_simulation_time = estimate_workload_time
                    best_configs = {"heavy": selected_heavy_action_frozen, "light": selected_light_actions}

                current_time = time.time()
                logging.debug(f"current global time: {(current_time - start_time)}")
                logging.debug(f"global time for indices: {idx_build_time}")

                light_root.update_statistics_with_mcts_reward(light_reward, selected_light_actions)
                # update the best gain for each query
                for sample_query_id in range(len(sampled_query_list)):
                    sample_query = sampled_query_list[sample_query_id]
                    run_time_of_sampled_query = run_time[sample_query_id]
                    if sample_query in best_micro_performance:
                        # if we get better improvement, then set it to better time
                        if run_time_of_sampled_query < best_micro_performance[sample_query]:
                            best_micro_performance[sample_query] = run_time_of_sampled_query
                    else:
                        best_micro_performance[sample_query] = run_time[sample_query_id]
                if sum(run_time) < best_reward:
                    best_reward = sum(run_time)
            # save the performance of current selected heavy action
            macro_performance_info[selected_heavy_action_frozen] = best_micro_performance

        logging.debug(f"macro_performance_info: {macro_performance_info}")
        # sys.stdout.flush()
        # ### obtain the best macro performances info
        best_slot_performance = dict()
        for selected_heavy_action_frozen, performance in macro_performance_info.items():
            for query, query_run_time in performance.items():
                if query in best_slot_performance:
                    if query_run_time < best_slot_performance[query]:
                        best_slot_performance[query] = query_run_time
                else:
                    best_slot_performance[query] = query_run_time
        logging.debug(f"best_slot_performance {best_slot_performance}")
        # after testing the performance of each configuration
        # generate the update information based on delta improvement
        update_info_slot = []
        for selected_heavy_actions in selected_heavy_action_batch:
            # generate intermediate result
            update_reward = []
            previous_performance = dict()
            for i in range(len(selected_heavy_actions)):
                if selected_heavy_actions[i] != terminate_action:
                    selected_heavy_action_frozen = frozenset(selected_heavy_actions[:i + 1])
                    logging.debug(f"current selected_heavy_action_frozen: {selected_heavy_action_frozen}")
                    applicable_query_performance = macro_performance_info[selected_heavy_action_frozen]
                    # generate reward based on the difference between previous performance and current performance
                    # the query for current indices
                    query_to_consider_new = index_to_applicable_queries[selected_heavy_actions[i]]
                    # get the query to consider
                    delta_improvement = 0
                    for query in query_to_consider_new:
                        previous_runtime = default_runtime[query]
                        current_runtime = default_runtime[query]
                        if query in applicable_query_performance:
                            current_runtime = applicable_query_performance[query]
                        if query in previous_performance:
                            previous_runtime = previous_performance[query]
                        delta_improvement += (previous_runtime - current_runtime)
                    previous_performance = applicable_query_performance
                    logging.debug(f"applicable_query_performance: {applicable_query_performance}")
                    logging.debug("previous_performance: {previous_performance}")
                    delta_improvement = max(delta_improvement, 0)
                    update_reward.append(delta_improvement)
            # update the tree based on the simulation results
            update_info_slot.append((update_reward, [heavy_action for heavy_action in selected_heavy_actions if
                                                     heavy_action != terminate_action]))
        logging.debug(f"update_info_slot: {update_info_slot}")
        heavy_root.update_batch(update_info_slot)
        previous_set = set(configuration_to_evaluate[evaluated_order[-1]])
        heavy_root.print_reward_info()
        end_episode_time = time.time()
        logging.debug(f"current time: {(end_episode_time - start_time)}")
        logging.debug(f"time for indices: {idx_build_time}")
        logging.info(f"current best heavy action: {heavy_root.best_actions()}")
        t1 += max_delay_time

    best_heavy_actions = heavy_root.best_actions()
    # best frozen heavy configs
    best_frozen_heavy_configs = frozenset(best_heavy_actions)
    if best_frozen_heavy_configs in light_tree_cache:
        light_root = light_tree_cache[best_frozen_heavy_configs]
    else:
        light_root = uct_node(round=0, tree_level=0, tree_height=light_tree_height, state=init_state, env=env,
                              space_type=SpaceType.Light)

    # additional step, tuning the final index configuration
    # build the indices
    add_action = set(best_heavy_actions) - previous_set
    drop_action = previous_set - set(best_heavy_actions)
    env.index_step(add_action, drop_action)
    micro_episode_final_tune = 50
    best_light_runtime = sum(default_runtime)
    for t2 in range(1, micro_episode_final_tune):
        # for the micro episode
        env.reset()
        selected_light_actions = light_root.sample(t1 * micro_episode + t2)
        # evaluate the light actions
        for selected_light_action in selected_light_actions:
            # move to next state
            state = env.step_without_evaluation(selected_light_action)
        # obtain run time info by running queries within timeout
        run_time = env.evaluate([query for query in range(nr_query)])
        # the total time of total queries
        total_run_time = sum(run_time)
        default_time = sum(default_runtime)
        if total_run_time < best_light_runtime:
            best_light_runtime = total_run_time
            best_light_config_simulation = selected_light_actions
        # the relative ration of the improvement, the less of total_run_time, the better
        light_reward = default_time / total_run_time
        logging.debug(f"light_action: {selected_light_actions}")
        logging.debug(f"light_reward: {light_reward}")
        light_root.update_statistics_with_mcts_reward(light_reward, selected_light_actions)
    best_light_actions = light_root.best_actions()
    # best heavy action from tree search
    logging.info(f"best configurations during the simulation {best_configs}")
    logging.info(f"best heavy action from tree search {heavy_root.best_actions()}")
    logging.info(f"best_light_actions {best_light_actions}")

    final_tune_time = time.time()
    logging.info(f"end: {final_tune_time}")

    logging.info(f"Summary: Total Tuning Time {(final_tune_time - start_tune_time) / 3600} hours")
    env.print_action_summary(best_heavy_actions + best_light_actions)



# -----------------------------------------------------------------------
# Copyright (c) 2021    Cornell Database Group
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
# IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
# -----------------------------------------------------------------------

import logging
import time

import gym
import udo_optimization

from udo.mcts.mcts_node import SpaceType
from udo.mcts.uct_node import uct_node

def run_simplifed_udo_agent(driver, queries, candidate_indices, tuning_config):
    start_tune_time = time.time()
    env = gym.make('udo_optimization-v0', driver=driver, queries=queries, candidate_indices=candidate_indices,
                   config=tuning_config)
    env.reset()
    duration_in_seconds = tuning_config['duration'] * 3600
    init_state = env.state_decoder(0)
    root = uct_node(0, 0, tuning_config['horizon'], init_state, env, space_type=SpaceType.All)

    all_queries = list(queries.values())
    start_time = time.time()
    current_time = time.time()

    ep = 0
    logging.debug(f"start: {time.time()}")
    while (current_time - start_time) < duration_in_seconds:
        # for the micro episode
        selected_light_actions = root.sample(ep)
        logging.debug(f"selected_light_actions: {selected_light_actions}")
        # evaluate the light actions
        env.reset()
        for selected_light_action in selected_light_actions:
            # move to next state
            state = env.step_without_evaluation(selected_light_action)
        # obtain run time info by running queries within timeout
        run_time = env.evaluate(all_queries)
        # the total time of sampled queries
        total_run_time = sum(run_time)
        # the default time of sampled queries
        default_time = sum(env.default_runtime)
        # the relative ration of the improvement, the less of total_run_time, the better
        # light_reward = default_time / total_run_time
        light_reward = max(default_time - total_run_time, 0)
        logging.debug(f"light reward: {light_reward}")
        root.update_statistics_with_mcts_reward(light_reward, selected_light_actions)
        root.print_reward_info()
        current_best = root.best_actions()
        logging.debug(f"current best action: {current_best}")
        logging.debug(f"runtime: {total_run_time}")
        # add more episode
        ep += 1
    final_tune_time = time.time()
    logging.debug(f"end: {final_tune_time}")
    # tuning summary
    best_actions = root.best_actions()
    logging.info(f"Summary: Total Tuning Time {(final_tune_time - start_tune_time) / 3600} hours")
    env.print_action_summary(best_actions)



# -----------------------------------------------------------------------
# Copyright (c) 2021    Cornell Database Group
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
# IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
# -----------------------------------------------------------------------

import logging

class OrderOptimizer:
    """
        Optimize the evaluation order using the estimated cost
    """

    def __init__(self, index_card_info):
        """cardinality information"""
        self.index_card_info = index_card_info

    def index_build_cost(self, current_index_list):
        """index build cost according to table cardinality"""
        return sum(self.index_card_info[current_index] for current_index in current_index_list)

    def greedy_min_cost_order(self, selected_action_batch):
        """determine the order to evaluate those actions of the given batch using greedy"""
        batch_size = len(selected_action_batch)
        if batch_size < 3:
            return range(batch_size)
        else:
            selected_action_batch_set = [set(batch) for batch in selected_action_batch]
            # take the first
            first_action = selected_action_batch_set[0]
            second_action = selected_action_batch_set[1]
            # cost for the first two elements
            common = first_action.intersection(second_action)
            cost = self.index_build_cost(first_action) + self.index_build_cost(second_action) - self.index_build_cost(
                common)
            order = [0, 1]
            for selected_action_idx in range(2, len(selected_action_batch_set)):
                # obtain select action
                selected_action = selected_action_batch_set[selected_action_idx]
                # consider to insert the new action to the first or last element
                current_cost1 = cost + self.index_build_cost(selected_action) - self.index_build_cost(
                    selected_action.intersection(selected_action_batch_set[order[0]]))
                current_cost2 = cost + self.index_build_cost(selected_action) - self.index_build_cost(
                    selected_action.intersection(selected_action_batch_set[order[-1]]))
                if current_cost2 > current_cost1:
                    min_cost = current_cost1
                    min_pos = -1
                else:
                    min_cost = current_cost2
                    min_pos = len(order) - 1
                # consider to insert the new action to the middle element
                for insert_pos in range(0, len(order) - 1):
                    prev_pos = order[insert_pos]
                    next_pos = order[insert_pos + 1]
                    current_cost = cost + self.index_build_cost(selected_action_batch_set[prev_pos].intersection(
                        selected_action_batch_set[next_pos])) - self.index_build_cost(
                        selected_action.intersection(selected_action_batch_set[prev_pos])) - self.index_build_cost(
                        selected_action.intersection(selected_action_batch_set[next_pos])) + self.index_build_cost(
                        selected_action)
                    if current_cost < min_cost:
                        min_cost = current_cost
                        min_pos = insert_pos
                # update the order
                order.insert(min_pos + 1, selected_action_idx)
            logging.debug("min cost order ", order)
            return order



# -----------------------------------------------------------------------
# Copyright (c) 2021    Cornell Database Group
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
# IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
# -----------------------------------------------------------------------

import logging
import math
import random
import time

import gym
import numpy as np
from gym import spaces

class UDOEnv(gym.Env):
    """the environment of DBMS optimizer"""

    def __init__(self, driver, queries, candidate_indices, config):
        """ init method
        Args:
            driver: DBMS connector
            queries: queries to optimize
            candidate_indices: candidate indices to select
            config: the configuration for tuning

        Return: the environment
        """
        super(UDOEnv, self).__init__()

        self.driver = driver

        # initial index action tuning space
        self.candidate_indices = candidate_indices
        self.index_candidate_num = len(self.candidate_indices)
        logging.debug(f"the total number of index candidates is {self.index_candidate_num}")

        # initial system parameter space
        self.parameter_candidate = self.driver.get_system_parameter_space()
        self.parameter_candidate_num = len(self.parameter_candidate)
        logging.debug(f"the total number of tuning system parameters is {self.parameter_candidate_num}")

        # combine the actions from 2 sub actions
        # action space
        self.nA_index = self.index_candidate_num
        self.nA_parameter = sum(self.parameter_candidate)
        self.nA = int(self.nA_index + self.nA_parameter)
        self.action_space = spaces.Discrete(self.nA)
        logging.debug(f"action space {self.nA}")

        # state space
        self.nS_index = int(math.pow(2, self.index_candidate_num))
        self.nS_parameter = np.prod(self.parameter_candidate)
        # self.nS = self.nS_index * self.nS_parameter
        logging.debug(f"index state space {self.nS_index}")
        logging.debug(f"parameter state space {self.nS_parameter}")

        # our transition matrix is a deterministic matrix
        # the observation space
        observation_space_array = np.concatenate([np.full(self.index_candidate_num, 1), self.parameter_candidate])
        self.observation_space = spaces.MultiDiscrete(observation_space_array)
        self.current_state = np.concatenate(
            [np.zeros(self.index_candidate_num), np.zeros(self.parameter_candidate_num)])

        # the MDP init setting
        # index build time

        # horizon
        self.horizon = config['horizon']
        self._step = 0
        # current step
        # for runtime statistics
        self.accumulated_index_time = 0
        self.start_time = time.time()

        # get all queries in the given benchmark
        self.queries = queries
        self.query_ids = list(self.queries.keys())
        self.nr_query = len(self.queries)
        self.query_sqls = [self.queries[query_id] for query_id in self.query_ids]
        query_to_id = {self.query_ids[idx]: idx for idx in range(self.nr_query)}
        self.index_to_applicable_queries = list(
            map(lambda x: list(map(lambda y: query_to_id[y], x[3])), self.candidate_indices))

        # the default run time
        default_time_out_per_query = 6
        # config['default_query_time_out']
        time_out_ratio = 1.1
        # config['time_out_ratio']
        input_runtime_out = [default_time_out_per_query] * self.nr_query
        self.default_runtime = self.driver.run_queries_with_timeout(self.query_sqls, input_runtime_out)
        self.runtime_out = [
            time_out_ratio * query_runtime if query_runtime < default_time_out_per_query
            else default_time_out_per_query for query_runtime in self.default_runtime]

        self.sample_rate = 1
        # config['sample_rate']
        self.best_state = None
        self.best_run_performance = sum(self.default_runtime)
        logging.info("start to tuning your database")
        logging.info(f"timeout for queries {self.runtime_out}")

    def state_decoder(self, num):
        """decode a vector to a state number"""
        index_pos = int(num % self.nS_index)
        index_state_string = np.binary_repr(int(index_pos), width=self.index_candidate_num)[::-1]
        # index stata represented in string
        index_state = list(map(int, index_state_string))
        parameter_value = int(num / self.nS_index)
        parameter_pos = [0] * self.parameter_candidate_num
        for i in range(self.parameter_candidate_num):
            parameter_pos[i] = (parameter_value % self.parameter_candidate[i])
            parameter_value = int(parameter_value / self.parameter_candidate[i])
        return index_state + parameter_pos

    def state_encoder(self, state):
        """encode a state number to a vector"""
        index_state = state[:self.index_candidate_num]
        parameter_state = state[self.index_candidate_num:]
        index_pos = int("".join([str(int(a)) for a in reversed(index_state)]), 2)
        parameter_pos = 0
        parameter_base = 1
        for i in range(len(self.parameter_candidate)):
            parameter_pos = parameter_pos + parameter_state[i] * parameter_base
            parameter_base = parameter_base * self.parameter_candidate[i]
        pos = index_pos + parameter_pos * self.nS_index
        return int(pos)

    def retrieve_heavy_actions(self, state):
        """obtain available heavy actions given a state"""
        index_state = state[:self.index_candidate_num]
        # check which indices are available or not
        candidate_index_action = [i for i in range(len(index_state)) if index_state[i] == 0]
        return candidate_index_action

    def retrieve_light_actions(self, state):
        """obtain available light actions given a state"""
        parameter_state = state[self.index_candidate_num:]
        candidate_parameter_action = []
        parameter_sum = 0
        for i in range(len(parameter_state)):
            # if the current parameter state is a default state, we switch its value
            if parameter_state[i] == 0:
                for j in range(1, self.parameter_candidate[i]):
                    candidate_parameter_action.append(self.nA_index + parameter_sum + j)
            parameter_sum += self.parameter_candidate[i]
        all_light_actions = candidate_parameter_action
        return all_light_actions

    def retrieve_actions(self, state):
        """retrieve available actions for a state including both light actions and heavy actions"""
        heavy_action = self.retrieve_heavy_actions(state)
        light_action = self.retrieve_light_actions(state)
        return heavy_action + light_action

    def transition(self, state, action):
        """transition from a state and an action to a next state"""
        assert action < self.nA
        index_state = state[:self.index_candidate_num]
        parameter_state = state[self.index_candidate_num:]
        if (action < self.nA_index):
            # action is the index action
            index_action = action
            index_state[index_action] = 1
        else:
            # action is the parameter action
            parameter_action = action - self.nA_index
            parameter_value = 0
            parameter_type = 0
            while parameter_type < len(self.parameter_candidate):
                parameter_range = self.parameter_candidate[parameter_type]
                # test whether cover this range
                if parameter_action < (parameter_value + parameter_range):
                    parameter_value = parameter_action - parameter_value
                    break
                parameter_value = parameter_value + parameter_range
                parameter_type += 1
            parameter_state[parameter_type] = parameter_value
        next_state = index_state + parameter_state
        return next_state, self.state_encoder(next_state)

    def step_without_evaluation(self, action):
        """move to the next state given the action and current state"""
        state = self.current_state
        # parameter state and action
        index_current_state = state[:self.index_candidate_num]
        parameter_current_state = state[self.index_candidate_num:]
        parameter_action = action - self.nA_index
        parameter_value = 0
        for parameter_type in range(len(self.parameter_candidate)):
            parameter_range = self.parameter_candidate[parameter_type]
            if parameter_action < (parameter_value + parameter_range):
                parameter_value = parameter_action - parameter_value
                break
            parameter_value = parameter_value + parameter_range
        parameter_current_state[parameter_type] = parameter_value
        next_state = np.concatenate([index_current_state, parameter_current_state])
        self.current_state = next_state
        return next_state

    def retrieve_light_action_command(self, action):
        parameter_action = action - self.nA_index
        parameter_value = 0
        parameter_type = 0
        for parameter_type in range(len(self.parameter_candidate)):
            parameter_range = self.parameter_candidate[parameter_type]
            if parameter_action < (parameter_value + parameter_range):
                parameter_value = parameter_action - parameter_value
                break
            parameter_value = parameter_value + parameter_range
        return self.driver.get_system_parameter_command(parameter_type, parameter_value)

    def evaluate(self, sampled_queries):
        """evaluate current state using the sampled queries"""
        state = self.current_state
        index_current_state = state[:self.index_candidate_num]
        parameter_current_state = state[self.index_candidate_num:]
        # change system parameter
        self.driver.change_system_parameter(parameter_current_state)
        # invoke queries
        run_time = self.driver.run_queries_with_timeout(
            [self.query_sqls[sampled_query] for sampled_query in sampled_queries],
            [self.runtime_out[sampled_query] for sampled_query in sampled_queries])
        logging.debug(f"evaluate time {sum(run_time)}")
        return run_time

    def index_step(self, add_actions, remove_actions):
        """index step"""
        for add_action in add_actions:
            index_to_create = self.candidate_indices[add_action]
            # build index action
            logging.debug(f"create index {index_to_create}")
            self.driver.build_index(index_to_create)
        for remove_action in remove_actions:
            index_to_drop = self.candidate_indices[remove_action]
            # drop index action
            logging.debug(f"drop index {index_to_drop}")
            self.driver.drop_index(index_to_drop)

    def index_add_step(self, add_action):
        """create indexes"""
        index_to_create = self.candidate_indices[add_action]
        # an index build action
        logging.debug(f"create index {index_to_create}")
        self.driver.build_index(index_to_create)

    def index_drop_step(self, remove_actions):
        """drop indexes"""
        for remove_action in remove_actions:
            index_to_drop = self.candidate_indices[remove_action]
            # drop index action
            logging.debug(f"drop index {index_to_drop}")
            self.driver.drop_index(index_to_drop)

    def step(self, action):
        """invoke an action and move to the next step"""
        state = self.current_state
        # parameter state and action
        index_current_state = state[:self.index_candidate_num]
        parameter_current_state = state[self.index_candidate_num:]
        if action < self.nA_index:
            # index action, create indices
            if index_current_state[action] == 0:
                index_start_time = time.time()
                self.index_add_step(action)
                index_end_time = time.time()
                index_time = (index_end_time - index_start_time)
                self.accumulated_index_time = self.accumulated_index_time + index_time
                index_current_state[action] = 1
        else:
            # parameter action, switch to different parameters
            parameter_action = action - self.nA_index
            parameter_value = 0
            for parameter_type in range(len(self.parameter_candidate)):
                parameter_range = self.parameter_candidate[parameter_type]
                if parameter_action < (parameter_value + parameter_range):
                    parameter_value = parameter_action - parameter_value
                    break
                parameter_value = parameter_value + parameter_range
            parameter_current_state[parameter_type] = parameter_value
            self.driver.change_system_parameter(parameter_current_state)

        # heavy actions, and we only consider queries applicable to selected indices
        active_indices = [index_pos for index_pos in range(self.nA_index) if index_current_state[index_pos] == 1]
        query_to_consider = set(
            [applicable_query for applicable_queries in
             list(map(lambda x: self.index_to_applicable_queries[x], active_indices)) for applicable_query in
             applicable_queries])

        if len(query_to_consider) == 0:
            query_to_consider = range(self.nr_query)

        # obtain sample number
        sample_num = math.ceil(self.sample_rate * len(query_to_consider))
        # generate sample queries
        sample_queries = random.sample(list(query_to_consider), k=sample_num)

        # invoke queries
        run_time = self.driver.run_queries_with_timeout(
            [self.query_sqls[sample_query] for sample_query in sample_queries],
            [self.runtime_out[sampled_query] for sampled_query in sample_queries])

        next_state = np.concatenate([index_current_state, parameter_current_state])
        self.current_state = next_state

        # scale the reward
        reward = sum(self.default_runtime) / sum(run_time)
        current_time = time.time()
        estimate_workload_time = sum(run_time) + sum(
            [self.default_runtime[query] for query in range(self.nr_query) if query not in sample_queries])

        logging.debug(f"run time: {run_time}")
        logging.debug(f"index time: {self.accumulated_index_time}")
        logging.debug(f"evaluate time: {sum(run_time)}")
        logging.debug(f"next state: {next_state}")
        logging.debug(f"tuning duration: {(current_time - self.start_time)}", )
        logging.debug(f"estimate whole workload time: {estimate_workload_time}")

        if estimate_workload_time < self.best_run_performance:
            self.best_run_performance = estimate_workload_time
            self.best_state = self.current_state

        self._step += 1
        if self._step == self.horizon:
            self._step = 0
            return next_state, reward, True, {}
        else:
            return next_state, reward, False, {}

    def reset(self):
        """reset the state"""
        if self.current_state is not None:
            state = self.current_state
            index_current_state = state[:self.index_candidate_num]
            parameter_current_state = state[self.index_candidate_num:]
            # drop all indices at the current state
            for i in range(self.index_candidate_num):
                if index_current_state[i] == 1:
                    index_drop_sql = self.candidate_indices[i]
                    self.driver.drop_index(index_drop_sql)
            # set the parameter to default value, the first value
            self.driver.change_system_parameter(np.zeros(self.parameter_candidate_num))
        self.current_state = np.concatenate(
            [np.zeros(self.index_candidate_num), np.zeros(self.parameter_candidate_num)])
        return self.current_state

    def print_state_summary(self, state):
        index_state = state[:self.index_candidate_num]
        parameter_state = state[self.index_candidate_num:]
        logging.info(f"Best index configurations:")
        for i in range(len(index_state)):
            if index_state[i] == 1:
                index_to_create = self.candidate_indices[i]
                self.driver.build_index_command(index_to_create)
        logging.info(f"Best system parameters:")
        for i in range(len(parameter_state)):
            logging.info(self.driver.get_system_parameter_command(i, parameter_state[i]))

    def print_action_summary(self, actions):
        """print action summary given actions"""
        logging.info(f"Best index configurations:")
        best_indices = [self.candidate_indices[action_idx] for action_idx in actions if action_idx < self.nA_index]
        best_sys_actions = [action_idx for action_idx in actions if action_idx >= self.nA_index]
        for best_index in best_indices:
            logging.info(self.driver.build_index_command(best_index))
        logging.info(f"Best system parameters:")
        for sys_action in best_sys_actions:
            logging.info(self.retrieve_light_action_command(sys_action))



# -----------------------------------------------------------------------
# Copyright (c) 2021    Cornell Database Group
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
# IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
# -----------------------------------------------------------------------

from enum import Enum


class SelectionPolicy(Enum):
    UCB1 = 1
    UCBV = 2


class UpdatePolicy(Enum):
    MCTS = 1
    RAVE = 2


class SpaceType(Enum):
    Light = 1
    Heavy = 2
    All = 3


class mcts_node(object):

    def select_action(self):
        """select actions"""
        return None

    def playout(self, current_level_action):
        """playout from a leaf node"""
        return None

    def sample(self, round):
        """sample from current node"""
        return None

    def opt_policy(self):
        """retrieve the current optimal policy"""
        return None

## MCTS



# -----------------------------------------------------------------------
# Copyright (c) 2021    Cornell Database Group
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
# IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
# -----------------------------------------------------------------------

class AbstractDriver(object):
    def __init__(self, driver_name, conf, sys_params):
        self.driver_name = driver_name
        self.config = conf
        self.sys_params = sys_params
        self.sys_params_type = len(self.sys_params)
        self.sys_params_space = [len(specific_parameter) for specific_parameter in self.sys_params]

    def __str__(self):
        return self.driver_name

    def connect(self):
        """connect to a DBMS"""
        raise NotImplementedError("%s does not implement connect function" % (self.driver_name))

    def run_queries_with_timeout(self, query_list, timeout):
        """run queries with specific timeout"""
        raise NotImplementedError("%s does not implement run queries with timeout function" % (self.driver_name))

    def run_queries_without_timeout(self, query_list):
        """run queries without specific timeout"""
        raise NotImplementedError("%s does not implement run queries without timeout function" % (self.driver_name))

    def build_index(self, index_creation_sql):
        """build index via SQL"""
        raise NotImplementedError("%s does not implement build index function" % (self.driver_name))

    def drop_index(self, index_drop_sql):
        """drop index via SQL"""
        raise NotImplementedError("%s does not implement drop index function" % (self.driver_name))

    def build_index_command(self, index_to_create):
        """build index command"""
        return None

    def set_system_parameter(self, parameter_sql):
        """switch system parameters via SQL"""
        raise NotImplementedError("%s does not implement set system parameter function" % (self.driver_name))

    def change_system_parameter(self, parameter_choices):
        """change system parameter values using the input parameter choices"""
        for i in range(self.sys_params_type):
            parameter_choice = int(parameter_choices[i])
            parameter_change_sql = self.sys_params[i][parameter_choice]
            self.set_system_parameter(parameter_change_sql)

    def get_system_parameter_command(self, parameter_type, parameter_value):
        """get the system parameter command based on the parameter_type and its value"""
        return self.sys_params[parameter_type][parameter_value]

    def get_system_parameter_space(self):
        """return the search space of system parameter"""
        return self.sys_params_space

## CLASS



from setuptools import setup, find_packages

setup(name='udo_db',
      version='0.18',
      description='a universal optimizer for database system',
      author='Junxiong Wang',
      author_email='chuangzhetianxia@gmail.coma',
      url='https://ovss.github.io/UDO/',
      download_url='https://github.com/OVSS/UDO/archive/refs/tags/0.01.tar.gz',
      keywords=['Database Optimization', 'OLAP', 'Index selection', 'System parameters'],
      packages=find_packages(),
      license='MIT',
      install_requires=[  # I get to this in a second
          'numpy==1.19.2',
          'h5py==2.10.0',
          'testresources>=2.0.1',
          'gym>=0.17.3',
          'udo_optimization>=0.0.2',
          'tensorflow>=2.4.1',
          'keras>=2.4.3',
          'keras-rl>=0.4.2',
          'mysqlclient>=2.0.3',
          'psycopg2>=2.8.6'
      ],
      classifiers=[
          'Development Status :: 3 - Alpha',
          # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
          'Intended Audience :: Developers',  # Define that your audience are developers
          'Topic :: Database',
          'License :: OSI Approved :: MIT License',  # Again, pick a license
          'Programming Language :: Python :: 3',  # Specify which python versions that you want to support
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
      ])



# -----------------------------------------------------------------------
# Copyright (c) 2021    Cornell Database Group
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
# IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
# -----------------------------------------------------------------------

import logging
import math
import random

from .mcts_node import *


class uct_node(mcts_node):
    def __init__(self, round, tree_level, tree_height, state, env, space_type=SpaceType.All,
                 selection_policy=SelectionPolicy.UCBV, update_policy=UpdatePolicy.RAVE, terminate_action=None):
        self.create_in = round
        # construct the transaction space, index space and parameter space
        self.env = env
        self.state = state
        if space_type is SpaceType.Light:
            self.actions = env.retrieve_light_actions(state)
        elif space_type is SpaceType.Heavy:
            self.actions = env.retrieve_heavy_actions(state)
        else:
            self.actions = env.retrieve_actions(state)
        self.nr_action = len(self.actions)
        self.tree_level = tree_level
        # for the children node
        self.children = [None] * self.nr_action
        # for the number of tries of children node
        self.nr_tries = [0] * self.nr_action
        self.first_moment = [0] * self.nr_action
        self.second_moment = [0] * self.nr_action
        self.total_visit = 0
        self.priority_actions = self.actions.copy()
        self.tree_height = tree_height
        self.terminate_action = terminate_action
        # update policy
        self.selection_policy = selection_policy
        self.update_policy = update_policy
        self.space_type = space_type
        # tuning parameters for UCB1
        self.explore_rate = 1.2
        # tuning parameters for UCBV
        self.bound = 0.1
        self.const = 0.1

    def select_action(self):
        """select action according to rl policy"""
        if len(self.priority_actions) > 0:
            selected_action = random.choice(self.priority_actions)
            self.priority_actions.remove(selected_action)
            selected_action_idx = self.actions.index(selected_action)
            return selected_action, selected_action_idx
        else:
            # select actions according to ucb1
            best_action_idx = -1
            best_ucb_score = -1
            for action_idx in range(self.nr_action):
                mean = self.first_moment[action_idx] / self.nr_tries[action_idx]
                variance = max((self.second_moment[action_idx] / self.nr_tries[action_idx] - (mean * mean)), 0)
                if self.selection_policy == SelectionPolicy.UCB1:
                    ucb_score = mean + self.explore_rate * math.sqrt(
                        math.log(self.total_visit) / self.nr_tries[action_idx])
                elif self.selection_policy == SelectionPolicy.UCBV:
                    ucb_score = mean + math.sqrt(
                        self.const * variance * math.log(self.total_visit) / self.nr_tries[action_idx]) + (
                                        3 * self.bound * math.log(self.total_visit) / self.nr_tries[action_idx])
                else:
                    raise ValueError('Selection policy is unavailable.')
                if ucb_score > best_ucb_score:
                    best_ucb_score = ucb_score
                    best_action_idx = action_idx
            return self.actions[best_action_idx], best_action_idx

    def playout(self, current_level_action):
        """randomly select other actions until reaching to terminate state"""
        current_level_state = self.state
        selected_action_path = []
        for current_level in range(self.tree_level + 1, self.tree_height):
            current_level_state, current_state_id = self.env.transition(current_level_state,
                                                                        current_level_action)
            if self.space_type is SpaceType.Light:
                current_candidate_actions = self.env.retrieve_light_actions(current_level_state)
            elif self.space_type is SpaceType.Heavy:
                current_candidate_actions = self.env.retrieve_heavy_actions(current_level_state)
            elif self.space_type == SpaceType.All:
                current_candidate_actions = self.env.retrieve_actions(current_level_state)
            # choose one random action from the action space
            if self.terminate_action is not None:
                current_candidate_actions.append(self.terminate_action)
            current_level_action = random.choice(current_candidate_actions)
            selected_action_path.append(current_level_action)
        return selected_action_path

    def sample(self, round):
        """sample actions from search tree"""
        if self.nr_action == 0:
            return []
        else:
            # inner node
            selected_action, selected_action_idx = self.select_action()
            if self.terminate_action is not None and selected_action == self.terminate_action:
                # for the terminate action, stop expansion
                return [selected_action]
            can_expand = (self.create_in != round) and (self.tree_level < self.tree_height)
            if can_expand and not self.children[selected_action_idx]:
                state, state_idx = self.env.transition(self.state, selected_action)
                self.children[selected_action_idx] = uct_node(round=0, tree_level=self.tree_level + 1,
                                                              tree_height=self.tree_height, state=state, env=self.env,
                                                              space_type=self.space_type,
                                                              selection_policy=self.selection_policy,
                                                              update_policy=self.update_policy,
                                                              terminate_action=self.terminate_action)
            child = self.children[selected_action_idx]
            # recursively sample the tree
            if child:
                return [selected_action] + child.sample(round)
            else:
                return [selected_action] + self.playout(selected_action)

    def update_statistics_with_mcts_reward(self, reward, selected_actions):
        """update mcts statistics using reward information"""
        rewards = {selected_action: reward for selected_action in selected_actions}
        self.update_statistics_with_delta_reward(rewards, selected_actions)

    def update_statistics_with_delta_reward(self, rewards, selected_actions):
        """update mcts statistics using intermediate reward information"""
        if self.update_policy is UpdatePolicy.RAVE:
            for action_idx in range(self.nr_action):
                current_action = self.actions[action_idx]
                # we use the RAVE update policy
                if current_action in selected_actions:
                    reward = rewards[current_action]
                    self.total_visit += 1
                    self.nr_tries[action_idx] += 1
                    self.first_moment[action_idx] += reward
                    self.second_moment[action_idx] += reward * reward
                    # update the reward for subtree
                    if self.children[action_idx] is not None:
                        next_selected_actions = list(filter(lambda x: x != current_action, selected_actions))
                        self.children[action_idx].update_statistics_with_delta_reward(rewards, next_selected_actions)
        else:
            current_action = selected_actions[self.tree_level]
            for action_idx in range(self.nr_action):
                if self.actions[action_idx] == current_action:
                    reward = rewards[current_action]
                    self.total_visit += 1
                    self.nr_tries[action_idx] += 1
                    self.first_moment[action_idx] += reward
                    self.second_moment[action_idx] += reward * reward
                    # self.mean_reward[action_idx] = self.first_moment[action_idx] / self.nr_tries[action_idx]
                    if self.children[action_idx] is not None:
                        self.children[action_idx].update_statistics_with_delta_reward(rewards, selected_actions)

    def update_batch(self, update_infos):
        for (rewards, select_actions) in update_infos:
            rewards = {select_actions[i]: rewards[i] for i in range(len(rewards))}
            self.update_statistics_with_delta_reward(rewards, select_actions)

    def print_reward_info(self):
        """print the reward information"""
        mean_reward = [0] * self.nr_action
        for i in range(self.nr_action):
            if self.nr_tries[i] > 0:
                mean_reward[i] = self.first_moment[i] / self.nr_tries[i]
            else:
                mean_reward[i] = 0
        logging.debug(f"first layer avg reward {mean_reward}")

    def best_actions(self):
        """best actions from the search tree"""
        best_mean = 0
        best_action_idx = -1
        for action_idx in range(self.nr_action):
            if self.nr_tries[action_idx] > 0:
                mean = self.first_moment[action_idx] / self.nr_tries[action_idx]
                if mean > best_mean:
                    best_mean = mean
                    best_action_idx = action_idx
        if self.children[best_action_idx] is not None:
            return [self.actions[best_action_idx]] + self.children[best_action_idx].best_actions()
        else:
            return [self.actions[best_action_idx]]



import logging

import gym
import udo_optimization

from rl.agents import SARSAAgent
from rl.policy import BoltzmannQPolicy
from tensorflow.keras.layers import Dense, Activation, Flatten
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam

def run_sarsa_agent(driver, queries, candidate_indices, tuning_config):
    # Get the environment and extract the number of actions.
    env = gym.make("udo_optimization-v0", driver=driver, queries=queries, candidate_indices=candidate_indices,
                   config=tuning_config)
    env.horizon = tuning_config['horizon']

    nb_actions = env.action_space.n
    logging.info(f"nr action: {nb_actions}")
    logging.info(f"observation space: {env.observation_space.shape}")

    # Next, we build a very simple model.
    model = Sequential()
    model.add(Flatten(input_shape=(1,) + env.observation_space.shape))
    model.add(Dense(52))
    model.add(Activation('relu'))
    model.add(Dense(252))
    model.add(Activation('relu'))
    model.add(Dense(526))
    model.add(Activation('relu'))
    model.add(Dense(252))
    model.add(Activation('relu'))
    model.add(Dense(nb_actions))
    model.add(Activation('linear'))

    logging.info(model.summary())

    # SARSA does not require a memory.
    policy = BoltzmannQPolicy()
    # policy.select_action()
    sarsa = SARSAAgent(model=model, nb_actions=nb_actions, nb_steps_warmup=10, policy=policy)
    sarsa.compile(Adam(lr=1e-3), metrics=['mae'])

    # Okay, now it's time to learn something! We visualize the training here for show, but this
    # slows down training quite a lot. You can always safely abort the training prematurely using
    # Ctrl + C.
    sarsa.fit(env, nb_steps=500, visualize=False, verbose=2)

    # After training is done, we save the final weights.
    # sarsa.save_weights('sarsa_{}_weights.h5f'.format(udo_optimization-v0), overwrite=True)

    # Finally, evaluate our algorithm for 5 episodes.
    sarsa.test(env, nb_episodes=5, visualize=False)
    env.print_state_summary(env.best_state)
